package tracker;
import tools.Logger;
import tools.LogLevel;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import protocolP2P.ProtocolP2PPacketTCP;
import protocolP2P.ProtocolP2PPacket;
import protocolP2P.RequestResponseCode;
import protocolP2P.Payload;
import protocolP2P.Register;
import protocolP2P.Unregister;
import tools.HostItem;
import java.util.ArrayList;
import java.util.List;
import java.io.IOException;
import exception.LocalException;
import java.util.Map;
import java.util.HashMap;
import protocolP2P.DiscoverRequest;
import protocolP2P.DiscoverResponse;
import protocolP2P.FileList;
import protocolP2P.HashRequest;
import localException.InternalError;


/** Tracker management implementation with tcp
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class TrackerManagementTCP implements Runnable {

	private int port;
	private Logger logger;
	private ServerSocket socket;
	private List<HostItem> hostList = new ArrayList<>();
	private Map<String, List<HostItem>> fileList = new HashMap<>();

	/** Constructor with port and logger.
	 * @param port	Port used to listen.
	 * @param logger	Logger object
	 */
	public TrackerManagementTCP(int port, Logger logger) {
		this.port = port;
		this.logger = logger;
		try {
			socket = new ServerSocket(port);
		} catch (SocketException e) {
			logger.writeTCP("Error: cannot listen on port " + port, LogLevel.Error);
			System.exit(-1);
		} catch (IOException e) {
			logger.writeTCP("Error: cannot openning socket", LogLevel.Error);
			System.exit(-2);
		}
	}

	/** Implementation of runnable. This methods allows to run the server.
	 */
	public void run() {
		logger.writeTCP("Tracker sucessfully started", LogLevel.Info);
		do {
			try {
				Socket s = socket.accept();
				ClientHandler c = new ClientHandler(s);
				(new Thread(c)).start();
			} catch (IOException e) {
				logger.writeTCP("Error while accepting new connection", LogLevel.Warning);
			}
		} while(true);
	}


	/** Private runnable class allowing to serve one client.
	 */
	private class ClientHandler implements Runnable {
		private Socket s;
		private String addr;
		/** Constructor with a socket.
		 * @param	s	Socket of this client
		 */
		public ClientHandler(Socket s) {
			this.s = s;
			this.addr = "[" +s.getInetAddress().getHostAddress() + "]:" + s.getPort() + " ";
		}

		/** Implementation of runnable. This method allow to serve one client.
		 */
		public void run() {

			boolean end = false;
			logger.writeTCP(addr + "New connection", LogLevel.Action);
			do {
				end = handleRequest();
			} while(!end);
			logger.writeTCP(addr + "End of connection", LogLevel.Action);
		}

		/** Respond to next request incomming on socket s.
		 * @param s	Socket used to read request and send response
		 * @return	true if cannot expect another request (ie, socket is closed)
		 */
		private boolean handleRequest() {
			try {
				ProtocolP2PPacketTCP pd = new ProtocolP2PPacketTCP((Object)s);
				Payload p = pd.getPayload();
				switch (p.getRequestResponseCode()) {
					case LOAD_REQUEST:
						logger.writeTCP("Received LOAD_REQUEST from host " + pd.getHostItem() + ", sending NOT_FOUND", LogLevel.Action);
						sendNotFound(pd);
						break;
					case LIST_REQUEST:
						logger.writeTCP("Received LIST_REQUEST from host " + pd.getHostItem() + ", sending EMPTY_DIRECTORY", LogLevel.Action);
						sendEmptyDirectory(pd);
						break;
					case HASH_REQUEST:
						logger.writeTCP("Received HASH_REQUEST from host " + pd.getHostItem() + ", sending NOT_FOUND", LogLevel.Action);
						sendNotFound(pd);
						break;
					case REGISTER:
						handleRegister(pd);
						break;
					case UNREGISTER:
						handleUnregister(pd);
						break;
					case DISCOVER_REQUEST:
						handleDiscover(pd);
						break;
					default:
						logger.writeTCP("Received grabbage from host " + pd.getHostItem(), LogLevel.Action);
						sendInternalError(pd);
						break;
				}
			} catch (IOException e) {
				logger.writeTCP(e, LogLevel.Warning);
				return true;
			} catch (LocalException e) {
				logger.writeTCP(e, LogLevel.Warning);
				return true;
			}
			return false;
		}
		
		/** Send a not found message.
		 * @param pd	ProtocolP2PPacketTCP to respond
		 */
		private void sendNotFound(ProtocolP2PPacketTCP pd) {
			try {
				pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.NOT_FOUND)));
			} catch (Exception e) {
				logger.writeTCP(e, LogLevel.Error);
			}
		}

		/** Send an empty directory message.
		 * @param pd	ProtocolP2PPacketTCP to respond
		 */
		private void sendEmptyDirectory(ProtocolP2PPacketTCP pd) {
			try {
				pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.EMPTY_DIRECTORY)));
			} catch (Exception e) {
				logger.writeTCP(e, LogLevel.Error);
			}
		}

		/** Send an internal error message.
		 * @param pd	ProtocolP2PPacketTCP to respond
		 */
		private void sendInternalError(ProtocolP2PPacketTCP pd) {
			logger.writeTCP("Internal Error", LogLevel.Warning);
			try {
				pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.INTERNAL_ERROR)));
			} catch (Exception e) {
				logger.writeTCP(e, LogLevel.Error);
			}
		}

		/** Handle Registering
		 * @param pd	ProtocolP2PPacketTCP to respond
		 * @throws	InternalError
		 */
		private void handleRegister(ProtocolP2PPacketTCP pd) throws InternalError {
			Payload p = pd.getPayload();
			assert p instanceof Register : "payload must be an instance of Register";
			if (!(p instanceof Register)) {
				throw new InternalError();
			} 
			// add host to known host list
			HostItem host = ((Register)p).getHostItem();
			if (!hostList.contains(host)) {
				hostList.add(host);
			}
			// send a list request
			try {
				ProtocolP2PPacket pLReq = (ProtocolP2PPacket) new ProtocolP2PPacketTCP(new Payload(RequestResponseCode.LIST_REQUEST));
				pLReq.sendRequest((Object)host.getTCPSocket());
				logger.writeTCP("Received REGISTER from host " + pd.getHostItem() + ". Adding host " + host + " to list. Sending List request", LogLevel.Action);
				handleListResponse((ProtocolP2PPacketTCP)pLReq.receiveResponse(), host);
				//host.closeTCPSocket();
			} catch (Exception e) {
				// remove from list because list request could not be send
				hostList.remove(host);
				logger.writeTCP("Aborting the add of host " + host, LogLevel.Action);
				logger.writeTCP(e, LogLevel.Error);
			}
		}

		/** Handle Unregistering
		 * @param pd	ProtocolP2PPacketTCP to respond
		 * @throws	InternalError
		 */
		private void handleUnregister(ProtocolP2PPacketTCP pd) throws InternalError {
			Payload p = pd.getPayload();
			assert p instanceof Unregister : "payload must be an instance of Unregister";
			if (!(p instanceof Unregister)) {
				sendInternalError(pd);
				throw new InternalError();
			}
			HostItem host = ((Unregister)p).getHostItem();
			logger.writeTCP("Received UNREGISTER from host " + pd.getHostItem() + ". Removing host " + host, LogLevel.Action);
			hostList.remove(host);
			for(String f: fileList.keySet()) {
				fileList.get(f).remove(host);
				if(fileList.get(f).isEmpty()) {
					fileList.remove(f);
				}
			}
		}

		/** Handle Discover request
		 * @param pd	ProtocolP2PPacketTCP to respond
		 * @throws	InternalError
		 */
		private void handleDiscover(ProtocolP2PPacketTCP pd) throws InternalError {
			logger.writeTCP("Received DISCOVER REQUEST from host " + pd.getHostItem(), LogLevel.Action);
			Payload p = pd.getPayload();
			assert p instanceof DiscoverRequest : "payload must be an instance of DiscoverRequest";
			if (!(p instanceof DiscoverRequest)) {
				sendInternalError(pd);
			} else {
				String filename = ((HashRequest)p).getFilename();
				try {
					pd.sendResponse((ProtocolP2PPacket)new ProtocolP2PPacketTCP((Payload)new DiscoverResponse(filename, fileList.getOrDefault(filename, hostList))));
				} catch (Exception e) {
					logger.writeTCP(e, LogLevel.Error);
				}
			}
		}

		/** Handle List Responses
		 * @param pd	ProtocolP2PPacketTCP response
		 * @throws	InternalError
		 */
		private void handleListResponse(ProtocolP2PPacketTCP pd, HostItem host) throws InternalError {
			logger.writeTCP("Received LIST RESPONSE from host " + host, LogLevel.Action);
			Payload p = pd.getPayload();
			assert p instanceof FileList: "payload must be an instance of FileList";
			if (!(p instanceof FileList)) {
				throw new InternalError();
			} else {
				String[] f = ((FileList)p).getFileList();
				for (String file: f) {
					List<HostItem> h = fileList.get(file);
					if (h != null) {
						if (!h.contains(host)) {
							h.add(host);
						}
					} else {
						List<HostItem> emptyH = new ArrayList<>();
						emptyH.add(host);
						fileList.put(file, emptyH);
					} 
				}
			}
		}
	}
}
